/*
 * Copyright (c) 2004, 2005, 2006, 2007, 2008
 *	Luiz Otavio O Souza <loos.br@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * $Id: pgsql.h,v 1.1.2.3 2008/05/16 00:46:50 loos-br Exp $
 */

#ifndef PGSQL_H
#define PGSQL_H

#ifndef PGSQLCONFDIR
#define PGSQLCONFDIR            "/usr/local/etc/msn-proxy/pgsql"
#endif
#ifndef PGSQLCONF
#define PGSQLCONF       	"conf"
#endif
#ifndef PGSQLTABLE
#define PGSQLTABLE      	"tables"
#endif
#ifndef DEFAULT_PGSQL_CONNINFO
#define DEFAULT_PGSQL_CONNINFO	"host=localhost dbname=msn-proxy port=5432 " \
				"user=msn-proxy password=secret"
#endif

#include "tree.h"
#include "string.h"
#include <libpq-fe.h>

struct table__ {
    RB_ENTRY(table__)	table_;
    string		name;
    string		sql;
    size_t		size;
    int			exist;
};

struct tables {
    struct table__	*rbh_root; /* root of the tree */
};

typedef struct {
    struct tables	table_head;	/* sql to create table */
    PGconn		*pgsql;		/* pgsql internal data */
    string		conninfo;	/* pgsql conninfo data */
    int			affected;	/* affected rows */
    int			connected;	/* is pgsql connected ? */
    void		(*info)(const char *, ...);
    void		(*debug)(const char *, ...);
} pgsql_;

int		db_read_conf(pgsql_ *pgsql, const char *configdir);
int		db_init(pgsql_ *pgsql, void (*debug)(const char *, ...),
				       void (*info)(const char *, ...));
int		db_query2(pgsql_ *pgsql, string *query);
char		*db_row(PGresult *res, int row, int column);
char		*db_error(pgsql_ *pgsql);
void		db_close(pgsql_ *pgsql);
PGresult	*db_free(PGresult *res);
PGresult 	*db_query(pgsql_ *pgsql, string *query);
__uint32_t	db_count(PGresult *res);

#endif
